import universe.envs.vnc_env
from universe.envs.vnc_env import VNCEnv
from universe.envs.vnc_core_env import GymCoreEnv, GymCoreSyncEnv
from universe.envs.vnc_flashgames import FlashgamesEnv
from universe.envs.vnc_internet import InternetEnv
from universe.envs.vnc_starcraft import StarCraftEnv
from universe.envs.vnc_gtav import GTAVEnv
from universe.envs.vnc_wog import WorldOfGooEnv
from universe.envs.dummy_vnc_env import DummyVNCEnv
